from yattag.simpledoc import dict_to_attrs, html_escape, attr_escape, SimpleDoc, DocError
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union
from typing import cast

try:
    range = xrange  # for Python 2/3 compatibility
except NameError:
    pass

__all__ = ['Doc']

class SimpleInput(object):

    """
    class representing text inputs, password inputs, hidden inputs etc...
    """
    
    def __init__(self, name, tpe, attrs):
        # type: (str, str, Dict[str, Union[str, int, float]]) -> None
        self.name = name
        self.tpe = tpe
        self.attrs = attrs
        
    def render(self, defaults, errors, error_wrapper, stag_end = ' />'):
        # type: (Dict[str, str], Dict[str, str], Tuple[str, str], str) -> str
        lst = [] # List[str]
        attrs = dict(self.attrs)
        error = errors and self.name in errors
        if error:
            _add_class(attrs, 'error')
            lst.append(error_wrapper[0])
            lst.append(html_escape(errors[self.name]))
            lst.append(error_wrapper[1])
                
        if self.name in defaults:
            if(self.tpe == 'file'):
                raise DocError('Default value for HTML form input of type "file" is not supported')

            attrs['value'] = str(defaults[self.name])
        attrs['name'] = self.name
        lst.append('<input type="%s" %s%s' % (self.tpe, dict_to_attrs(attrs), stag_end))
        
        return ''.join(lst)
        

class CheckableInput(object):
    tpe = 'checkbox'

    def __init__(self, name, attrs):
        # type: (str, Dict[str, Union[str, int, float]]) -> None
        self.name = name
        self.rank = 0
        self.attrs = attrs
        
    def setrank(self, n):
        # type: (int) -> None
        self.rank = n
    
    @classmethod
    def match(cls, default, value):
        # type: (Any, Union[str, int, float]) -> bool
        if isinstance(default, str):
            return value == default
        elif isinstance(default, (tuple, list, set)):
            return value in default
        return False
     
    def checked(self, defaults):
        # type: (Dict[str, Union[List[str], str]]) -> bool
        try:
            default = defaults[self.name]
        except KeyError:
            return False
        try:
            value = self.attrs['value']
        except KeyError:
            return False
        return self.__class__.match(default, value)    
        
    
    def render(self, defaults, errors, error_wrapper, stag_end = ' />'):
        # type: (Dict[str, Union[List[str], str]], Any, Tuple[str, str], str) -> str
        lst = []
        attrs = dict(self.attrs)
        if self.rank == 0:
            if errors and self.name in errors:
                lst.append(error_wrapper[0])
                lst.append(html_escape(errors[self.name]))
                lst.append(error_wrapper[1])
                _add_class(attrs, 'error')
        
        if self.checked(defaults):
            attrs['checked'] = 'checked'
                
        attrs['name'] = self.name
        
        lst.append('<input type="%s" %s%s' % (self.__class__.tpe, dict_to_attrs(attrs), stag_end))

        return ''.join(lst)
        

class CheckboxInput(CheckableInput):
    pass
    
class RadioInput(CheckableInput):
    tpe = 'radio'
    
    @classmethod
    def match(cls, default, value):
        # type: (Any, Union[str, int, float]) -> bool
        if isinstance(default, str):
            return value == default
        return False

def groupclass(inputclass):
    # type: (Any) -> Any

    class InputGroup(object):

        def __init__(self, name):
            # type: (str) -> None
            self.name = name
            self.n_items = 0
            
        def input(self, attrs):
            # type: (Dict[str, Union[str, int, float]]) -> Any
            input_instance = inputclass(self.name, attrs)
            input_instance.setrank(self.n_items)
            self.n_items += 1
            return input_instance

    return InputGroup

class ContainerTag(object):

    tag_name = 'textarea' 

    def __init__(self, name, attrs):
        # type: (str, Dict[str, Union[str, int, float]]) -> None
        self.name = name
        self.attrs = attrs
        
    def render(self, defaults, errors, error_wrapper, inner_content = ''):
        # type: (Dict[str, str], Dict[str, str], Tuple[str, str], str) -> str
        lst = []
        attrs = dict(self.attrs)
        if errors and self.name in errors:
            lst.append(error_wrapper[0])
            lst.append(html_escape(errors[self.name]))
            lst.append(error_wrapper[1])
            _add_class(attrs, 'error')
        attrs['name'] = self.name

        lst.append('<%s %s>' % (self.__class__.tag_name, dict_to_attrs(attrs)))
        if self.name in defaults:
            lst.append(html_escape(str(defaults[self.name])))
        else:
            lst.append(inner_content)
        
        lst.append('</%s>' % self.__class__.tag_name)

        return ''.join(lst)

class Textarea(ContainerTag):
    pass

class Select(ContainerTag):
    tag_name = 'select'


class Option(object):
    def __init__(self, name, multiple, value, attrs):
        # type: (str, str, Union[str, int, float], Dict[str, Any]) -> None
        self.name = name
        self.multiple = multiple
        self.value = value
        self.attrs = attrs

    def render(self, defaults, errors, inner_content):
        # type: (Dict[str, str], Dict[str, str], str) -> str
        selected = False        
        if self.name in defaults:
            if self.multiple:
                if self.value in defaults[self.name]:
                    selected = True
            else:
                if self.value == defaults[self.name]:
                    selected = True
        lst = ['<option value="', attr_escape(self.value), '"']
        if selected:
            lst.append(' selected="selected"')
        if self.attrs:
            lst.append(' ')
            lst.append(dict_to_attrs(self.attrs))
        lst.append('>')
        lst.append(inner_content)
        lst.append('</option>')
        return ''.join(lst)
        
def _attrs_from_args(required_keys, *args, **kwargs):
    # type: (Any, Any, Union[str, int, float]) -> List[Any]
    # need to do all this to allow specifying attributes as (key, value) pairs
    # while maintaining backward compatibility with previous versions
    # of yattag, which allowed 'name', 'type', and 'value' attributes
    # as positional or as keyword arguments
    def raise_exception(arg):
        # type: (Any) -> None
        raise ValueError(
            "Optional attributes should be passed as (key, value) pairs or as keyword arguments."
            "Got %s (type %s)" % (repr(arg), repr(type(arg)))   
        )
    limit = 0
    for arg in args:
        if isinstance(arg, tuple):
            break 
        else:
            limit += 1
    if limit > len(required_keys):
        raise_exception(args[limit-1])
    attrs = dict(zip(required_keys[:limit],args[:limit]))
    for arg in args[limit:]:
        if isinstance(arg, tuple):
            attrs[arg[0]] = arg[1]
        else:
            raise_exception(arg)
    attrs.update(
        (('class', value) if key == 'klass' else (key, value))
        for key, value in kwargs.items()
    )

    required_attrs = []

    for key in required_keys:
        try:
            required_attrs.append(attrs.pop(key))
        except KeyError:
            raise ValueError(
                "the %s attribute is missing" % repr(key)
            )
    return required_attrs + [attrs]
   
class Doc(SimpleDoc):
    """
    The Doc class extends the SimpleDoc class with form rendering capabilities. 
    Pass default values or errors as dictionnaries to the Doc constructor, and 
    use the `input`, `textarea`, `select`, `option` methods
    to append form elements to the document.
    """
    
    SimpleInput = SimpleInput
    CheckboxInput = CheckboxInput
    RadioInput = RadioInput
    Textarea = Textarea
    Select = Select
    Option = Option
    
    class TextareaTag(object):
        def __init__(self, doc, name, attrs):
            # type: (Doc, str, Dict[str, Union[str, int, float]]) -> None
            # name is the name attribute of the textarea, ex: 'contact_message'
            # for <textarea name="contact_message">
            self.doc = doc
            self.name = name
            self.attrs = attrs
        
        def __enter__(self):
            # type: () -> None
            self.parent_tag = self.doc.current_tag
            self.doc.current_tag = self
            self.position = len(self.doc.result)
            self.doc._append('')
            
        def __exit__(self, tpe, value, traceback):
            # type: (Any, Any, Any) -> None
            if value is None:
                inner_content = ''.join(self.doc.result[self.position+1:])
                del self.doc.result[self.position+1:]              
                rendered_textarea = self.doc.__class__.Textarea(self.name, self.attrs).render(
                    defaults = self.doc.defaults,
                    errors = self.doc.errors,
                    inner_content = inner_content,
                    error_wrapper = self.doc.error_wrapper
                )
                self.doc.result[self.position] = rendered_textarea
                self.doc.current_tag = self.parent_tag
                
    
    class SelectTag(object):
        def __init__(self, doc, name, attrs):
            # type: (Doc, str, Dict[str, Union[str, int, float]]) -> None
            # name is the name attribute of the select, ex: 'color'
            # for <select name="color">
            self.doc = doc
            self.name = name
            self.attrs = attrs
            self.multiple = bool(attrs.get('multiple'))
            self.old_current_select = None
                
        def __enter__(self):
            # type: () -> None
            self.parent_tag = self.doc.current_tag
            self.doc.current_tag = self
            self.position = len(self.doc.result)
            self.doc._append('')
            self.old_current_select = self.doc.current_select
            self.doc.current_select = self
            
        def __exit__(self, tpe, value, traceback):
            # type: (Any, Any, Any) -> None
            if value is None:
                inner_content = ''.join(self.doc.result[self.position+1:])
                del self.doc.result[self.position+1:]
                rendered_select = self.doc.__class__.Select(self.name, self.attrs).render(
                    defaults = {},  # no defaults for the <select> tag. Defaults are handled by the <option> tags directly.
                    errors = self.doc.errors,
                    inner_content = inner_content,
                    error_wrapper = self.doc.error_wrapper
                )
                self.doc.result[self.position] = rendered_select
                self.doc.current_tag = self.parent_tag  
                self.doc.current_select = self.old_current_select
                

    class OptionTag(object):
        def __init__(self, doc, select, value, attrs):
            # type: (Doc, Doc.SelectTag, str, Dict[str, Union[str, int, float]]) -> None
            self.doc = doc
            self.select = select
            self.attrs = attrs
            self.value = value

        def __enter__(self):
            # type: () -> None
            self.parent_tag = self.doc.current_tag
            self.doc.current_tag = self
            self.position = len(self.doc.result)
            self.doc._append('')
            
        def __exit__(self, tpe, value, traceback):
            # type: (Any, Any, Any) -> None
            if value is None:
                inner_content = ''.join(self.doc.result[self.position+1:])
                del self.doc.result[self.position+1:]
                self.doc.result[self.position] = self.doc.__class__.Option(
                    name = self.select.name,
                    multiple = self.select.multiple,
                    value = self.value,
                    attrs = self.attrs
                ).render(
                    defaults = self.doc.defaults,
                    errors = self.doc.errors,
                    inner_content = inner_content
                )
                self.doc.current_tag = self.parent_tag

    
    def __init__(self, defaults = None, errors = None,
     error_wrapper = ('<span class="error">', '</span>'), *args, **kwargs):
        # type: (Optional[Dict[str, str]], Optional[Dict[str, str]], Tuple[str, str], Any, Any) -> None
        """
        creates a Doc instance
        
        defaults::
            optional dictionnary of values used to fill html forms
        errors::
            optional dictionnary of errors used to fill html forms
        
        Example 1::
            doc = Doc()
        
        Example 2::
            doc = Doc(
                defaults = {
                    'beverage': 'coffee',
                    'preferences': ['milk', 'sugar'],
                    'use_discount': True
                },
                errors = {
                    'preferences': "We ran out of milk!"
                }
            )
            
        Note: very often you'll want to call the `tagtext` method just after
        creating a Doc instance. Like this::
        
        doc, tag, text = Doc(defaults = {'color': 'blue'}).tagtext()
        
        This way, you can write `tag` (resp. `text`) in place of `doc.tag`
        (resp. `doc.text`). When writing long html templates or xml documents,
        it's a gain in readability and performance.
        """
                
        super(Doc, self).__init__(*args, **kwargs)
        self.defaults = defaults or {}
        self.errors = errors or {}
        self.error_wrapper = error_wrapper
        self.radios = {} # type: Dict[str, Any]
        self.checkboxes = {} # type: Dict[str, Any]
        self.current_select = None # type: Optional[Any]
        self.radio_group_class = groupclass(self.__class__.RadioInput)
        self.checkbox_group_class = groupclass(self.__class__.CheckboxInput)
        self._fields = set() # type: Set[Any]
        self._detached_errors_pos = [] # type: List[Any]
    
     
    def input(self, *args, **kwargs):
        # type: (Any, Union[str, int, float]) -> None
        "required attributes: 'name' and 'type'"
        name, type, attrs = _attrs_from_args(('name', 'type'), *args, **kwargs)
        self._fields.add(name)
        if type in (
            'text','file','tel', 'password', 'hidden', 'search', 'email', 'url', 'number',
            'range', 'date', 'datetime', 'datetime-local', 'month', 'week',
            'time', 'color'
        ): 
            self.asis(
                self.__class__.SimpleInput(name, type, attrs).render(
                    self.defaults, self.errors, self.error_wrapper, self._stag_end
                )
            )
            return
        if type == 'radio':
            if name not in self.radios:
                self.radios[name] = self.radio_group_class(name)
            checkable_group = self.radios[name]
        elif type == 'checkbox':
            if name not in self.checkboxes:
                self.checkboxes[name] = self.checkbox_group_class(name)
            checkable_group = self.checkboxes[name]
        else:
            if type == 'submit':
                raise DocError("Unhandled input type: submit. Use doc.stag('input', type = 'submit', value='whatever') instead.")
            else:
                raise DocError("Unknown input type: %s" % type)
        
        self._append(checkable_group.input(attrs).render(self.defaults, self.errors, self.error_wrapper, self._stag_end))
        
    def textarea(self, *args, **kwargs):
        # type: (Any, Union[str, int, float]) -> Doc.TextareaTag
        "required attribute: 'name'"
        name, attrs = _attrs_from_args(('name',), *args, **kwargs)
        self._fields.add(name)
        return self.__class__.TextareaTag(self, name, attrs)
        
    def select(self, *args, **kwargs):
        # type: (Any, Union[str, int, float]) -> Doc.SelectTag
        "required attribute: 'name'"
        name, attrs = _attrs_from_args(('name',), *args, **kwargs)
        self._fields.add(name)
        return self.__class__.SelectTag(self, name, attrs)
        
    def option(self, *args, **kwargs):
        # type: (Any, Union[str, int, float]) -> Doc.OptionTag
        "required attribute: 'value'"
        if self.current_select:
            value, attrs = _attrs_from_args(('value',), *args, **kwargs)
            return self.__class__.OptionTag(self, self.current_select, value, attrs)
        else:
            raise DocError("No <select> tag opened. Can't put an <option> here.")
            
    def detached_errors(self, render_function = None):
        # type: (Any) -> None
        self._detached_errors_pos.append((len(self.result), render_function or self.error_dict_to_string))
        self.result.append('')
        
    def error_dict_to_string(self, dct):
        # type: (Dict[str, str]) -> str
        if dct:
            doc, tag, text = SimpleDoc().tagtext()
            with tag('ul', klass='error-list'):
                for error in dct.values():
                    with tag('li'):
                        text(error)
            return doc.getvalue()
        else:
            return ''
                        
            
    def getvalue(self):
        # type: () -> str
        """
        returns the whole document as a string
        """
        for position, render_function in self._detached_errors_pos:
            self.result[position] = render_function(
                dict((name, self.errors[name]) for name in self.errors if name not in self._fields)
            )
        return ''.join(self.result)

def _add_class(dct, klass):
    # type: (Dict[str, Any], str) -> None
    classes = dct.get('class', '').split()
    if klass not in classes:
        dct['class'] = ' '.join(classes + [klass])



__all__ = ['SimpleDoc']

import re
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple
from typing import Union
from typing import cast

class DocError(Exception):
    pass

class SimpleDoc(object):

    """
    class generating xml/html documents using context managers

    doc, tag, text = SimpleDoc().tagtext()

    with tag('html'):
        with tag('body', id = 'hello'):
            with tag('h1'):
                text('Hello world!')

    print(doc.getvalue())
    """

    class Tag(object):
        def __init__(self, doc, name, attrs): # name is the tag name (ex: 'div')
            # type: (SimpleDoc, str, Dict[str, Union[str, int, float]]) -> None

            self.doc = doc
            self.name = name
            self.attrs = attrs

        def __enter__(self):
            # type: () -> None
            self.parent_tag = self.doc.current_tag
            self.doc.current_tag = self
            self.position = len(self.doc.result)
            self.doc._append('')

        def __exit__(self, tpe, value, traceback):
            # type: (Any, Any, Any) -> None
            if value is None:
                if self.attrs:
                    self.doc.result[self.position] = "<%s %s>" % (
                        self.name,
                        dict_to_attrs(self.attrs),
                    )
                else:
                    self.doc.result[self.position] = "<%s>" % self.name
                self.doc._append("</%s>" % self.name)
                self.doc.current_tag = self.parent_tag

    class DocumentRoot(object):

        class DocumentRootError(DocError, AttributeError):
            # Raising an AttributeError on __getattr__ instead of just a DocError makes it compatible
            # with the pickle module (some users asked for pickling of SimpleDoc instances).
            # I also keep the DocError from earlier versions to avoid possible compatibility issues
            # with existing code.
            pass

        def __getattr__(self, item):
            # type: (str) -> Any
            raise SimpleDoc.DocumentRoot.DocumentRootError("DocumentRoot here. You can't access anything here.")

    _newline_rgx = re.compile(r'\r?\n')

    def __init__(self, stag_end = ' />', nl2br = False):
        # type: (str, bool) -> None
        r"""
            stag_end:
                the string terminating self closing tags.
                This determines what tags produced using the `stag` method will look like.
                For example, if you set `stag_end='>'`, then `doc.stag('hr')` will
                produce a `<hr>` tag.
                If you set `stag_end=' />'` (the default), then `doc.stag('hr')` would
                instead produce a `<hr />` tag.
                If you set `nl2br=True`, then the `text` method will also
                produce `<br>` or `<br />` tags according to this preference when encountering
                new lines.
                Defaults to ' />'.

            nl2br:
                if set to True, the `text` method will turn new lines
                ('\n' or '\r\n' sequences) in the input to `<br />` html tags,
                or possibly to `<br>` tags if using the `stag_end` parameter to this effect.
                (see explanations about `stag_end` above).
                Defaults to False (new lines are not replaced).

        """
        self.result = [] # type: List[str]
        self.current_tag = self.__class__.DocumentRoot() # type: Any
        self._append = self.result.append
        assert stag_end in (' />', '/>', '>')
        self._stag_end = stag_end
        self._br = '<br' + stag_end
        self._nl2br = nl2br

    def tag(self, tag_name, *args, **kwargs):
        # type: (str, Tuple[str, Union[str, int, float]], Union[str, int, float]) -> Tag
        """
        opens a HTML/XML tag for use inside a `with` statement
        the tag is closed when leaving the `with` block
        HTML/XML attributes can be supplied as keyword arguments,
        or alternatively as (key, value) pairs.
        The values of the keyword arguments should be strings.
        They are escaped for use as HTML attributes
        (the " character is replaced with &quot;)

        In order to supply a "class" html attributes, you must supply a `klass` keyword
        argument. This is because `class` is a reserved python keyword so you can't use it
        outside of a class definition.

        Example::

            with tag('h1', id = 'main-title'):
                text("Hello world!")

            # <h1 id="main-title">Hello world!</h1> was appended to the document

            with tag('td',
                ('data-search', 'lemon'),
                ('data-order', '1384'),
                id = '16'
            ):
                text('Citrus Limon')

            # you get: <td data-search="lemon" data-order="1384" id="16">Citrus Limon</td>

        """
        return self.__class__.Tag(self, tag_name, _attributes(args, kwargs))


    def text(self, *strgs):
        # type: (str) -> None
        r"""
        appends 0 or more strings to the document
        the strings are escaped for use as text in html documents, that is,
        & becomes &amp;
        < becomes &lt;
        > becomes &gt;

        Example::

            username = 'Max'
            text('Hello ', username, '!') # appends "Hello Max!" to the current node
            text('16 > 4') # appends "16 &gt; 4" to the current node

        New lines ('\n' or '\r\n' sequences) are left intact, unless you have set the
        nl2br option to True when creating the SimpleDoc instance. Then they would be
        replaced with `<br />` tags (or `<br>` tags if using the `stag_end` option
        of the SimpleDoc constructor as shown in the example below).

        Example::

            >>> doc = SimpleDoc()
            >>> doc.text('pistachio\nice cream')
            >>> doc.getvalue()
            'pistachio\nice cream'

            >>> doc = SimpleDoc(nl2br=True)
            >>> doc.text('pistachio\nice cream')
            >>> doc.getvalue()
            'pistachio<br />ice cream'

            >>> doc = SimpleDoc(nl2br=True, stag_end='>')
            >>> doc.text('pistachio\nice cream')
            >>> doc.getvalue()
            'pistachio<br>ice cream'

        """
        for strg in strgs:
            transformed_string = html_escape(strg)
            if self._nl2br:
                self._append(
                    self.__class__._newline_rgx.sub(
                        self._br,
                        transformed_string
                    )
                )
            else:
                self._append(transformed_string)

    def line(self, tag_name, text_content, *args, **kwargs):
        # type: (str, str, Tuple[str, Union[str, int, float]], Union[str, int, float]) -> None
        """
        Shortcut to write tag nodes that contain only text.
        For example, in order to obtain::

            <h1>The 7 secrets of catchy titles</h1>

        you would write::

            line('h1', 'The 7 secrets of catchy titles')

        which is just a shortcut for::

            with tag('h1'):
                text('The 7 secrets of catchy titles')

        The first argument is the tag name, the second argument
        is the text content of the node.
        The optional arguments after that are interpreted as xml/html
        attributes. in the same way as with the `tag` method.

        Example::

            line('a', 'Who are we?', href = '/about-us.html')

        produces::

            <a href="/about-us.html">Who are we?</a>
        """
        with self.tag(tag_name, *args, **kwargs):
            self.text(text_content)

    def asis(self, *strgs):
        # type: (str) -> None
        """
        appends 0 or more strings to the documents
        contrary to the `text` method, the strings are appended "as is"
        &, < and > are NOT escaped

        Example::

            doc.asis('<!DOCTYPE html>') # appends <!DOCTYPE html> to the document
        """
        for strg in strgs:
            if strg is None:
                raise TypeError("Expected a string, got None instead.")
                # passing None by mistake was frequent enough to justify a check
                # see https://github.com/leforestier/yattag/issues/20
            self._append(strg)

    def nl(self):
        # type: () -> None
        self._append('\n')

    def attr(self, *args, **kwargs):
        # type: (Tuple[str, Union[str, int, float]], Union[str, int, float]) -> None
        """
        sets HTML/XML attribute(s) on the current tag
        HTML/XML attributes are supplied as (key, value) pairs of strings,
        or as keyword arguments.
        The values of the keyword arguments should be strings.
        They are escaped for use as HTML attributes
        (the " character is replaced with &quot;)
        Note that, instead, you can set html/xml attributes by passing them as
        keyword arguments to the `tag` method.

        In order to supply a "class" html attributes, you can either pass
        a ('class', 'my_value') pair, or supply a `klass` keyword argument
        (this is because `class` is a reserved python keyword so you can't use it
        outside of a class definition).

        Examples::

            with tag('h1'):
                text('Welcome!')
                doc.attr(id = 'welcome-message', klass = 'main-title')

            # you get: <h1 id="welcome-message" class="main-title">Welcome!</h1>

            with tag('td'):
                text('Citrus Limon')
                doc.attr(
                    ('data-search', 'lemon'),
                    ('data-order', '1384')
                )


            # you get: <td data-search="lemon" data-order="1384">Citrus Limon</td>

        """
        self.current_tag.attrs.update(_attributes(args, kwargs))

    def data(self, *args, **kwargs):
        # type: (Tuple[str, Union[str, int, float]], Union[str, int, float]) -> None
        """
        sets HTML/XML data attribute(s) on the current tag
        HTML/XML data attributes are supplied as (key, value) pairs of strings,
        or as keyword arguments.
        The values of the keyword arguments should be strings.
        They are escaped for use as HTML attributes
        (the " character is replaced with &quot;)
        Note that, instead, you can set html/xml data attributes by passing them as
        keyword arguments to the `tag` method.

        Examples::

            with tag('h1'):
                text('Welcome!')
                doc.data(msg='welcome-message')

            # you get: <h1 data-msg="welcome-message">Welcome!</h1>

            with tag('td'):
                text('Citrus Limon')
                doc.data(
                    ('search', 'lemon'),
                    ('order', '1384')
                )


            # you get: <td data-search="lemon" data-order="1384">Citrus Limon</td>

        """
        self.attr(
            *(('data-%s' % key, value) for (key, value) in args),
            **dict(('data-%s' % key, value) for (key, value) in kwargs.items())
        )

    def stag(self, tag_name, *args, **kwargs):
        # type: (str, Tuple[str, Union[str, int, float]], Union[str, int, float]) -> None
        """
        appends a self closing tag to the document
        html/xml attributes can be supplied as keyword arguments,
        or alternatively as (key, value) pairs.
        The values of the keyword arguments should be strings.
        They are escaped for use as HTML attributes
        (the " character is replaced with &quot;)

        Example::

            doc.stag('img', src = '/salmon-plays-piano.jpg')
            # appends <img src="/salmon-plays-piano.jpg" /> to the document

        If you want to produce self closing tags without the ending slash (HTML5 style),
        use the stag_end parameter of the SimpleDoc constructor at the creation of the
        SimpleDoc instance.

        Example::

            >>> doc = SimpleDoc(stag_end = '>')
            >>> doc.stag('br')
            >>> doc.getvalue()
            '<br>'
        """
        if args or kwargs:
            self._append("<%s %s%s" % (
                tag_name,
                dict_to_attrs(_attributes(args, kwargs)),
                self._stag_end
            ))
        else:
            self._append("<%s%s" % (tag_name, self._stag_end))

    def cdata(self, strg, safe = False):
        # type: (str, bool) -> None
        """
        appends a CDATA section containing the supplied string

        You don't have to worry about potential ']]>' sequences that would terminate
        the CDATA section. They are replaced with ']]]]><![CDATA[>'.

        If you're sure your string does not contain ']]>', you can pass `safe = True`.
        If you do that, your string won't be searched for ']]>' sequences.
        """
        self._append('<![CDATA[')
        if safe:
            self._append(strg)
        else:
            self._append(strg.replace(']]>', ']]]]><![CDATA[>'))
        self._append(']]>')

    def getvalue(self):
        # type: () -> str
        """
        returns the whole document as a single string
        """
        return ''.join(self.result)

    def tagtext(self):
        # type: () -> Tuple[SimpleDoc, Any, Any]
        """
        return a triplet composed of::
            . the document itself
            . its tag method
            . its text method

        Example::

            doc, tag, text = SimpleDoc().tagtext()

            with tag('h1'):
                text('Hello world!')

            print(doc.getvalue()) # prints <h1>Hello world!</h1>
        """

        return self, self.tag, self.text

    def ttl(self):
        # type: () -> Tuple[SimpleDoc, Any, Any, Any]
        """
        returns a quadruplet composed of::
            . the document itself
            . its tag method
            . its text method
            . its line method

        Example::

            doc, tag, text, line = SimpleDoc().ttl()

            with tag('ul', id='grocery-list'):
                line('li', 'Tomato sauce', klass="priority")
                line('li', 'Salt')
                line('li', 'Pepper')

            print(doc.getvalue())
        """
        return self, self.tag, self.text, self.line

    def add_class(self, *classes):
        # type: (str) -> None
        """
        adds one or many elements to the html "class" attribute of the current tag
        Example::
            user_logged_in = False
            with tag('a', href="/nuclear-device", klass = 'small'):
                if not user_logged_in:
                    doc.add_class('restricted-area')
                text("Our new product")

            print(doc.getvalue())

            # prints <a class="restricted-area small" href="/nuclear-device"></a>
        """
        self._set_classes(
            self._get_classes().union(classes)
        )

    def discard_class(self, *classes):
        # type: (str) -> None
        """
        remove one or many elements from the html "class" attribute of the current
        tag if they are present (do nothing if they are absent)
        """
        self._set_classes(
            self._get_classes().difference(classes)
        )

    def toggle_class(self, elem, active):
        # type: (str, bool) -> None
        """
        if active is a truthy value, ensure elem is present inside the html
        "class" attribute of the current tag, otherwise (if active is falsy)
        ensure elem is absent
        """
        classes = self._get_classes()
        if active:
            classes.add(elem)
        else:
            classes.discard(elem)
        self._set_classes(classes)


    def _get_classes(self):
        # type: () -> Set[str]
        try:
            current_classes = self.current_tag.attrs['class']
        except KeyError:
            return set()
        else:
            return set(current_classes.split())

    def _set_classes(self, classes_set):
        # type: (Set[str]) -> None
        if classes_set:
            self.current_tag.attrs['class'] = ' '.join(classes_set)
        else:
            try:
                del self.current_tag.attrs['class']
            except KeyError:
                pass

def html_escape(s):
    # type: (Union[str, int, float]) -> str
    if isinstance(s,(int,float)):
        return str(s)
    try:
        return s.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
    except AttributeError:
        raise TypeError(
            "You can only insert a string, an int or a float inside a xml/html text node. "
            "Got %s (type %s) instead." % (repr(s), repr(type(s)))
        )


def attr_escape(s):
    # type: (Union[str, int, float]) -> str
    if isinstance(s,(int,float)):
        return str(s)
    try:
        return s.replace("&", "&amp;").replace("<", "&lt;").replace('"', "&quot;")
    except AttributeError:
        raise TypeError(
            "xml/html attributes should be passed as strings, ints or floats. "
            "Got %s (type %s) instead." % (repr(s), repr(type(s)))
        )


ATTR_NO_VALUE = object()

def dict_to_attrs(dct):
    # type: (Dict[str, Any]) -> str
    return ' '.join(
        (key if value is ATTR_NO_VALUE
        else '%s="%s"' % (key, attr_escape(value)))
        for key,value in dct.items()
    )

def _attributes(args, kwargs):
    # type: (Any, Any) -> Dict[str, Any]
    lst = [] # type: List[Any]
    for arg in args:
        if isinstance(arg, tuple):
            lst.append(arg)
        elif isinstance(arg, str):
            lst.append((arg, ATTR_NO_VALUE))
        else:
            raise ValueError(
                "Couldn't make a XML or HTML attribute/value pair out of %s."
                % repr(arg)
            )
    result = dict(lst)
    result.update(
        (('class', value) if key == 'klass' else (key, value))
        for key,value in kwargs.items()
    )
    return result
